"""
Core datatypes involved in testing and impact analysis system
"""

import json
import uuid
from abc import ABC, abstractmethod
from typing import List

import attr
from pydash import py_


class Task(ABC):
    """
    This is basically X for any machine learning model. When this object
    contains a not None `tags` property, that is considered as it's y.
    """

    tags = attr.ib(default=None)
    is_gold: bool = attr.ib(default=False)

    @property
    @abstractmethod
    def id(self):
        """
        Unique identifier for this case
        """
        ...

    def __eq__(self, other):
        return self.id == other.id


class DictTask(Task, dict):
    """
    Dictionary task wrapping around any form of data.
    """

    id: str = attr.ib()

    @staticmethod
    def from_dict(d, id):
        task = DictTask(d)
        task.id = id
        return task


@attr.s(slots=True)
class SimulatedTurn:
    """
    One turn of each simulated call.

    Fields here have similar meaning as in the conversations table in
    production database.
    """

    id: str = attr.ib()
    type: str = attr.ib()
    sub_type: str = attr.ib()
    text: str = attr.ib()
    prediction = attr.ib()

    @staticmethod
    def from_dict(d, make_prediction=None):
        if make_prediction:
            pred = make_prediction(d["text"])
        else:
            if isinstance(d["prediction"], dict):
                pred = d["prediction"]
            else:
                pred = json.loads(d["prediction"])

        return SimulatedTurn(
            id=d["id"],
            type=d["type"],
            sub_type=d["sub_type"],
            text=d["text"],
            prediction=pred,
        )


@attr.s(slots=True)
class SimulatedCallTask(Task):
    """
    Task representing a single simulated call coming from plute's user
    simulator scripts.
    make_prediction is a function which generates prediction out of text
    """

    id: str = attr.ib()
    turns: List[SimulatedTurn] = attr.ib()

    @staticmethod
    def from_dict(d, make_prediction=None):
        return SimulatedCallTask(
            id=d["id"],
            turns=[
                SimulatedTurn.from_dict(td, make_prediction=make_prediction)
                for td in d["turns"]
                # NOTE: We remove bot turns (type = RESPONSE) for now. If you want
                #       that too for training models, note the following points:
                #       1. the simulated bot texts might not be the one used in production
                #       2. we don't get bot turns in plute production as of now.
                #       One you have considered both these cases, you can go ahead
                #       with a parser for bot turns too.
                if ("type" not in td) or (td["type"] == "INPUT")
            ],
        )


@attr.s(slots=True)
class CallTranscriptionTurn:
    """
    One turn of each agent-user transcribed call.

    Fields here have similar meaning as in the conversations table in
    production database.
    """

    id: str = attr.ib()
    type: str = attr.ib()
    text: str = attr.ib()

    @staticmethod
    def from_dict(d):

        return CallTranscriptionTurn(id=d["id"], type=d["type"], text=d["text"])


@attr.s(slots=True)
class CallTranscriptionTask(Task):
    """
    Task for using agent-user transcribed calls.
    """

    id: str = attr.ib()
    turns: List[CallTranscriptionTurn] = attr.ib()

    @staticmethod
    def from_dict(d, id):
        return CallTranscriptionTask(
            id=d["id"] if "id" in d else id,
            turns=[
                CallTranscriptionTurn.from_dict(td)
                for td in d["turns"]
                # NOTE: We remove bot turns (type = RESPONSE) for now. If you want
                #       that too for training models, note the following points:
                #       1. the simulated bot texts might not be the one used in production
                #       2. we don't get bot turns in plute production as of now.
                #       One you have considered both these cases, you can go ahead
                #       with a parser for bot turns too.
                if ("type" not in td) or (td["type"] == "INPUT")
            ],
        )


@attr.s(slots=True)
class AudioSegmentTask(Task):
    """
    Task for audio segment tagging like for VAD, diarization etc.
    """

    conversation_id: int = attr.ib()
    audio_url: str = attr.ib()

    @property
    def id(self):
        return self.conversation_id

    @staticmethod
    def from_dict(d):
        it = py_.pick(d, ["conversation_id", "audio_url"])
        return AudioSegmentTask(**it)


@attr.s(slots=True)
class ConversationTask(Task):
    """
    Plute style conversation (turn) task coming from tog.
    """

    alternatives = attr.ib()
    data_id = attr.ib()
    audio_url: str = attr.ib()
    call_uuid: str = attr.ib()
    conversation_uuid: str = attr.ib()
    state: str = attr.ib()
    reftime: str = attr.ib()
    prediction = attr.ib(default=None)
    raw = attr.ib(default=None)

    @property
    def id(self):
        return self.conversation_uuid

    @staticmethod
    def from_dict(d):
        call_uuid = (
            d.get("call_uuid") if d.get("call_uuid") is not None else d.get("call_id")
        )
        conversation_uuid = (
            d.get("conversation_uuid")
            if d.get("conversation_uuid") is not None
            else d.get("conversation_id")
        )
        if call_uuid is None or conversation_uuid is None:
            raise ValueError(f"No reference for call or conversation. {d.keys()}")
        if d.get("alternatives"):
            d["alternatives"] = json.dumps(d["alternatives"], ensure_ascii=False)
        return ConversationTask(
            **{
                **py_.pick(
                    d,
                    [
                        "alternatives",
                        "audio_url",
                        "state",
                        "reftime",
                        "prediction",
                    ],
                ),
                "data_id": str(conversation_uuid),
                "raw": d,
                "call_uuid": str(call_uuid),
                "conversation_uuid": str(conversation_uuid),
            }
        )


@attr.s(slots=True)
class DataGenerationTask(Task):
    """
    Data Generation Task.
    Direct intent-entity recording job.
    """

    id: str = attr.ib()

    @staticmethod
    def from_dict(d):
        return DataGenerationTask(id=d.get("id", uuid.uuid4().hex))



"""
Module for working with tog database
"""

import json
import os
import sqlite3
from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Union

import psycopg2
import pytz
from skit_fixdf.fix import datetime as fix_dt
from skit_labels import constants as const

from skit_labels.types import (
    AudioSegmentTask,
    CallTranscriptionTask,
    ConversationTask,
    DataGenerationTask,
    DictTask,
    SimulatedCallTask,
    Task,
)


def build_task(
    d: Dict, task_type: str, data_id: Optional[str] = None, tz=pytz.UTC
) -> Task:
    """
    Create a task from given data dictionary.
    """

    if task_type == "conversation":
        task = ConversationTask.from_dict(d)

        # Since the reftime from db is in UTC, we convert it to our timezone. This
        # is needed as saying 12 pm means different things in different timezones
        # and can't be translated without doing something stupid.
        task.reftime = fix_dt.to_datetime(task.reftime)
        task.reftime = task.reftime.astimezone(tz).isoformat()
    elif task_type == "simulated_call":
        task = SimulatedCallTask.from_dict(d)
    elif task_type == "audio_segment":
        task = AudioSegmentTask.from_dict(d)
    elif task_type == "dict":
        task = DictTask.from_dict(d, data_id)
    elif task_type == "call_transcription":
        task = CallTranscriptionTask.from_dict(d, data_id)
    elif task_type == "data_generation":
        task = DataGenerationTask.from_dict(d)
    else:
        raise TypeError(f"Invalid task type {task_type} provided.")
    return task


class SqliteDatabase:
    """
    Class mapping to a local sqlite database file which can keep only one job.
    """

    def __init__(self, filepath: str):
        self.filepath = filepath
        self._initialize()

    def _initialize(self):
        self.conn = sqlite3.connect(self.filepath)
        c = self.conn.cursor()
        c.execute(
            """CREATE TABLE IF NOT EXISTS data (
            data_id INTEGER NOT NULL,
            data TEXT NOT NULL,
            tag TEXT NOT NULL,
            is_gold BOOLEAN NOT NULL,
            tagged_time TEXT
        )"""
        )
        self.conn.commit()

    def insert_rows(self, rows: List):
        """
        Write rows in database. Each item of row is a tuple of following elements:
        - data_id : int
        - data: Dict
        - tag: Dict
        - is_gold: bool
        - tagged_time: Optional[str]
        """

        c = self.conn.cursor()
        c.executemany(
            "INSERT INTO data (data_id, data, tag, is_gold, tagged_time) VALUES (?, ?, ?, ?, ?)",
            [(i, json.dumps(d), json.dumps(t), g, tt) for i, d, t, g, tt in rows],
        )
        self.conn.commit()


class Database:
    """
    Class holding connection with backend database.
    """

    def __init__(
        self,
        db: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[Union[int, str]] = None,
    ):
        self._initialize(db=db, user=user, password=password, host=host, port=port)

    def _initialize(
        self,
        db: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[Union[int, str]] = None,
    ):
        db = db or os.getenv(const.TOGDB_DB, "tog")
        user = user or os.getenv(const.TOGDB_USER)
        password = password or os.getenv(const.TOGDB_PASSWORD)
        host = host or os.getenv(const.TOGDB_HOST, "localhost")
        port = port or os.getenv(const.TOGDB_PORT, "5432")

        if password is None:
            raise ValueError(
                "Credentials for Tog database not set. Check for missing environment variables."
            )

        self.conn = psycopg2.connect(
            host=host, database=db, user=user, password=password, port=port
        )

    def list_jobs(self) -> List[Dict]:
        with self.conn.cursor() as cur:
            # NOTE: We are not picking out task_type field since that
            #       collides with our task type names. Ideally we need to
            #       settle on same nomenclature.
            cur.execute(
                "SELECT id, name, description, config, language FROM jobs_job WHERE is_active"
            )

            jobs: List[Dict] = []
            for row in cur.fetchall():
                jobs.append(
                    {
                        "id": row[0],
                        "name": row[1],
                        "description": row[2],
                        "config": row[3],
                        "language": row[4],
                    }
                )

            return jobs


class AbstractJob(ABC):
    """
    Abstract class representing connection to a tagging job.
    """

    @abstractmethod
    def total(self, untagged=False) -> int:
        ...

    @abstractmethod
    def get_by_data_id(self, id: int, cache=True):
        ...

    @abstractmethod
    def get(self, untagged=False, itersize=1000, only_gold=False):
        ...


class Job(AbstractJob):
    """
    A Tog job which specifies a kind of tagging data set and problem.
    """

    def __init__(
        self,
        id: int,
        task_type="conversation",
        database: Optional[str] = None,
        tz=pytz.UTC,
        db: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[Union[str, int]] = None,
    ):
        self.id = id
        # TODO: Check task validity right here
        self.task_type = task_type

        self.db = database or Database(
            db=db, user=user, password=password, host=host, port=port
        )
        self._fetch_details()

        # Cache for keeping rows of job indexed by data_ids
        self.cache = {}
        self.tz = tz

    def _fetch_details(self):
        """
        Save metadata about the current job id
        """

        with self.db.conn.cursor() as cur:
            # NOTE: We are not picking out task_type field since that
            #       collides with our task type names. Ideally we need to
            #       settle on same nomenclature.
            cur.execute(
                f"SELECT name, description, config, language FROM jobs_job WHERE id = {self.id}"
            )
            try:
                self.name, self.description, self.config, self.lang = cur.fetchone()
            except TypeError:
                raise ValueError("Invalid job id")

    def total(self, untagged=False):
        """
        Return total number of items for this job. If `untagged` is True, consider
        untagged items in counting too.
        """

        with self.db.conn.cursor() as cur:
            cur.execute(
                f"SELECT count(*) FROM jobs_task WHERE job_id = {self.id} {'' if untagged else 'AND tag IS NOT NULL'}"
            )
            n = cur.fetchone()[0]
        return n

    def get_by_data_id(self, id: int, cache=True):
        """
        Return task and tag using the data id
        """

        if id in self.cache:
            return self.cache[id]

        with self.db.conn.cursor() as cur:
            cur.execute(
                f"""SELECT
              jobs_data.data, jobs_task.tag, jobs_task.is_gold, jobs_task.tagged_time
            FROM jobs_task INNER JOIN jobs_data ON
              jobs_data.id = jobs_task.data_id
            WHERE jobs_task.job_id = {self.id} AND jobs_data.data_id = '{id}'
            """
            )

            try:
                task_dict, tag_list, is_gold, tagged_time = cur.fetchone()
            except TypeError:
                raise RuntimeError("No item found for given data id")

            task = build_task(task_dict, self.task_type, tz=self.tz)
            task.is_gold = bool(is_gold)
            tag = json.loads(tag_list)

            if cache:
                self.cache[id] = (task, tag, tagged_time)
            return task, tag, tagged_time

    def type(self):
        """
        List all available job_ids
        """
        with self.db.conn.cursor(name="data_cursor") as cursor:
            cursor.execute('select "taskType" from jobs_job where id = %s', (self.id,))
            record = cursor.fetchone()
            if record:
                return record[0]
            raise ValueError(f"Invalid job id {self.id}")

    def get(
        self,
        untagged=False,
        itersize=1000,
        only_gold=False,
        start_date=None,
        end_date=None,
    ):
        """
        Return (generator) tagged tasks and tags from the database. Itersize sets
        the iteration size for the server sided cursor.

        If `untagged` is True, also return untagged items. This might be useful
        for checking, say, production metrics. If `only_gold` is True, return
        only items which are marked as gold.
        """
        query = f"""
        SELECT
            jobs_data.data,
            jobs_task.tag,
            jobs_task.is_gold,
            jobs_task.tagged_time,
            jobs_data.id
        FROM jobs_task INNER JOIN jobs_data ON
            jobs_data.id = jobs_task.data_id
        WHERE
            jobs_task.job_id = {self.id}
            {'' if untagged else 'AND jobs_task.tag IS NOT NULL'}
            {"AND jobs_task.is_gold = true" if only_gold else ''}
            {f"AND jobs_data.created_at >= '{start_date}'" if isinstance(start_date, str) else ''}
            {f"AND jobs_data.created_at < '{end_date}'" if isinstance(end_date, str) else ''}
        """

        with self.db.conn.cursor(name="data_cursor") as cur:
            cur.itersize = itersize
            cur.execute(query)

            for row in cur:
                task_dict, tag, is_gold, tagged_time, data_id = row
                task = build_task(task_dict, self.task_type, data_id, tz=self.tz)
                task.is_gold = bool(is_gold)
                yield task, tag, tagged_time


class JobLocal(AbstractJob):
    """
    A tog job relying on local sqlite database.
    """

    def __init__(self, filepath: str, task_type="conversation", tz=pytz.UTC):
        self.task_type = task_type
        self.conn = sqlite3.connect(filepath)
        self.cache = {}
        self.tz = tz

    def total(self, untagged=False):
        """
        Return total number of items for this job. If `untagged` is True, consider
        untagged items in counting too.
        """

        cur = self.conn.cursor()
        cur.execute(
            f"SELECT count(*) FROM data {'' if untagged else 'WHERE tag IS NOT NULL'}"
        )
        return cur.fetchone()[0]

    def get_by_data_id(self, id: int, cache=True, show_source=False):
        """
        Return task and tag using the data id
        NOTE: We are not using cache here.
        """

        on_source = ", source" if show_source else ""

        cur = self.conn.cursor()
        cur.execute(
            f"""SELECT
          data, tag, is_gold, tagged_time{on_source}
        FROM data
        WHERE data_id = '{id}'
        """
        )

        try:
            if show_source:
                task_dict, tag_list, is_gold, tagged_time, source = cur.fetchone()
            else:
                task_dict, tag_list, is_gold, tagged_time = cur.fetchone()
        except TypeError:
            raise RuntimeError("No item found for given data id")

        task = build_task(json.loads(task_dict), self.task_type, tz=self.tz)
        task.is_gold = bool(is_gold)
        tag = json.loads(tag_list)
        if show_source:
            return task, tag, tagged_time, source
        return task, tag, tagged_time

    def get(
        self,
        untagged=False,
        itersize=1000,
        only_gold=False,
        show_source=False,
        show_ids=False,
    ):
        """
        Return (generator) tagged tasks and tags from the database.

        If `untagged` is True, also return untagged items. This might be useful
        for checking, say, production metrics. If `only_gold` is True, return
        only items which are marked as gold.
        """
        cur = self.conn.cursor()
        on_source = ", source" if show_source else ""
        on_ids = "data_id," if show_ids else ""
        cur.execute(
            f"""SELECT
          {on_ids} data, tag, is_gold, tagged_time {on_source}
        FROM data
        {'' if untagged else 'WHERE tag IS NOT NULL'}
        {'AND is_gold = 1' if only_gold else ''}
        """
        )

        for row in cur:
            if show_source and show_ids:
                id_, task_dict, tag, is_gold, tagged_time, source = row
            elif show_source:
                task_dict, tag, is_gold, tagged_time, source = row
            elif show_ids:
                id_, task_dict, tag, is_gold, tagged_time = row
            else:
                task_dict, tag, is_gold, tagged_time = row
            task = build_task(json.loads(task_dict), self.task_type, tz=self.tz)
            task.is_gold = bool(is_gold)
            if show_source and show_ids:
                yield id_, task, json.loads(tag), tagged_time, source
            elif show_source:
                yield task, json.loads(tag), tagged_time, source
            elif show_ids:
                yield id_, task, json.loads(tag), tagged_time
            else:
                yield task, json.loads(tag), tagged_time



"""
Command line interface for interacting with a tog datasets.
"""
import argparse
from ast import arg
import asyncio
import os
import sys
from datetime import datetime

import pytz

from skit_labels import __version__, commands
from skit_labels import constants as const
from skit_labels import utils


def is_timezone(value: str) -> str:
    if value not in pytz.all_timezones:
        raise argparse.ArgumentTypeError(
            f"Unknown timezone {value}. Lookup `pytz.all_timezones`."
        )
    return value


def is_numeric(value: str) -> str:
    if not isinstance(value, str):
        raise argparse.ArgumentTypeError(f"{value} is not a string.")
    if not value.isdigit():
        raise argparse.ArgumentTypeError(f"{value} is not a numeric value.")
    return value


def date_type(value: str):
    try:
        return datetime.strptime(value, "%Y-%m-%d").date()
    except ValueError:
        raise argparse.ArgumentTypeError(f"Invalid date {value}, expected YYYY-MM-DD.")


def create_job_args(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    parser.add_argument(
        "-j",
        "--job-id",
        type=is_numeric,
        required=True,
        help="Id of the tog dataset that we want to download.",
    )
    parser.add_argument(
        "--db", type=str, help="Database name.", default=os.environ.get(const.TOGDB_DB)
    )
    parser.add_argument(
        "--host",
        type=str,
        help="Database host.",
        default=os.environ.get(const.TOGDB_HOST),
    )
    parser.add_argument(
        "--port",
        type=int,
        help="Database port.",
        default=os.environ.get(const.TOGDB_PORT),
    )
    parser.add_argument(
        "--user",
        type=str,
        help="Database user.",
        default=os.environ.get(const.TOGDB_USER),
    )
    parser.add_argument(
        "--password",
        type=str,
        help="Database password.",
        default=os.environ.get(const.TOGDB_PASSWORD),
    )
    return parser


def add_job_args(fn):
    def wrapper(*args, **kwargs):
        parser = fn(*args, **kwargs)
        return create_job_args(parser)

    return wrapper


@add_job_args
def build_dataset_from_tog_command(
    parser: argparse.ArgumentParser,
) -> argparse.ArgumentParser:
    parser.add_argument(
        "-o",
        "--output-format",
        type=str,
        help="Store dataset in supported formats.",
        choices=[".csv", ".sqlite"],
        default=".csv",
    )
    parser.add_argument(
        "-tz",
        "--timezone",
        type=is_timezone,
        help="Timezone to parse datetime values. Like 'America/Los_Angeles', 'Asia/Kolkata' etc.",
        default=pytz.UTC,
    )
    parser.add_argument(
        "--batch-size",
        type=int,
        default=500,
        help="Number of items to download in a batch.",
    )
    parser.add_argument(
        "--full",
        action="store_true",
        help="If provided, download all data instead of including untagged datapoints.",
    )
    parser.add_argument(
        "-tt",
        "--task-type",
        type=str,
        default=const.TASK_TYPE__CONVERSATION,
        help="Task type for deserialization.",
        choices=const.TASK_TYPES,
    )
    parser.add_argument(
        "--start-date",
        type=date_type,
        help="Filter items added to the dataset after this date. (inclusive)",
    )
    parser.add_argument(
        "--end-date",
        type=date_type,
        help="Filter items added to the dataset before this date. (exclusive)",
    )
    return parser


def build_dataset_from_dvc_command(
    parser: argparse.ArgumentParser,
) -> argparse.ArgumentParser:
    parser.add_argument(
        "--repo", type=str, required=True, help="DVC enabled git repository."
    )
    parser.add_argument("--path", type=str, required=True, help="Path to the dataset.")
    parser.add_argument(
        "--remote",
        type=str,
        help="Remote. Required only if the repo "
        "hasn't set a default remote. This is usually a bucket name.",
    )
    return parser


def build_download_command(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    data_source_parsers = parser.add_subparsers(dest="data_source")
    build_dataset_from_tog_command(
        data_source_parsers.add_parser(
            const.SOURCE__DB,
            help="Download a dataset of a given id from the database.",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
    )
    build_dataset_from_dvc_command(
        data_source_parsers.add_parser(
            const.SOURCE__DVC,
            help="Download a dataset from a dvc enabled repo.",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
    )
    return parser


def upload_dataset_to_tog_command(
    parser: argparse.ArgumentParser,
) -> argparse.ArgumentParser:
    parser.add_argument(
        "--url",
        type=str,
        help="URL of the dataset server. Optionally set the DATASET_SERVER_URL environment variable.",
        default=os.environ.get(const.DATASET_SERVER_URL),
    )
    parser.add_argument(
        "--token",
        type=str,
        help="The organization authentication token.",
        default=utils.read_session(),
    )
    parser.add_argument(
        "-i",
        "--input",
        type=str,
        help="The raw data to be uploaded.",
    )
    return parser


def build_upload_command(parser: argparse.ArgumentParser) -> None:
    data_source_parsers = parser.add_subparsers(dest="data_source")
    upload_dataset_to_tog_command(
        data_source_parsers.add_parser(
            const.SOURCE__DB,
            help="Upload a dataset to the database. Creates a new dataset if dataset id not provided",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
    )


def build_cli():
    parser = argparse.ArgumentParser(
        description=f"skit-labels {__version__}. Command line interface for interacting with labelled datasets.",
    )
    parser.add_argument(
        "-v", action="count", help="Increase verbosity.", dest="verbosity", default=0
    )
    command_parsers = parser.add_subparsers(dest="command")
    build_download_command(
        command_parsers.add_parser(
            const.DOWNLOAD,
            help="Download a dataset. of a given id from the database.",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
    )
    build_upload_command(
        command_parsers.add_parser(
            const.UPLOAD,
            help="Upload a dataset.",
        )
    )
    create_job_args(
        command_parsers.add_parser(
            const.DESCRIBE, help="Describe a dataset for a given tog dataset id."
        )
    )
    create_job_args(
        command_parsers.add_parser(
            const.STATS, help="Get tagged/untagged points for a given tog dataset id."
        )
    )
    return parser


def cmd_to_str(args: argparse.Namespace) -> str:
    utils.configure_logger(args.verbosity)
    if args.command == const.DOWNLOAD and args.data_source == const.SOURCE__DB:
        return commands.download_dataset_from_db(
            args.job_id,
            args.task_type,
            args.timezone,
            full=args.full,
            batch_size=args.batch_size,
            output_format=args.output_format,
            start_date=args.start_date,
            end_date=args.end_date,
            db=args.db,
            host=args.host,
            port=args.port,
            user=args.user,
            password=args.password,
        )
    elif args.command == const.DOWNLOAD and args.data_source == const.SOURCE__DVC:
        return commands.download_dataset_from_dvc(args.repo, args.path, args.remote)
    elif args.command == const.UPLOAD and args.data_source == const.SOURCE__DB:
        if not args.token:
            raise ValueError(
                "Token is required for uploading to the database."
                "Use [skit-auth](https://github.com/skit-ai/skit-auth) to obtain the token."
            )

        if args.input is None:
            is_pipe = not os.isatty(sys.stdin.fileno())
            if is_pipe:
                args.input = sys.stdin.readline().strip()
            else:
                raise argparse.ArgumentTypeError(
                    "Expected to receive --input=<file> or its valued piped in."
                )

        errors, df_size = asyncio.run(
            commands.upload_dataset_to_db(
                args.input,
                args.url,
                args.token,
                job_id=args.job_id,
            )
        )

        if errors:
            error_summary = "\n".join(set(errors))
            return (
                f"Encountered {len(errors)} over {df_size}.\nSummary:\n{error_summary}."
            )
    elif args.command == const.DESCRIBE:
        return commands.describe_dataset(
            args.job_id,
            db=args.db,
            host=args.host,
            port=args.port,
            user=args.user,
            password=args.password,
        )
    elif args.command == const.STATS:
        return commands.stat_dataset(
            args.job_id,
            db=args.db,
            host=args.host,
            port=args.port,
            user=args.user,
            password=args.password,
        )


def main():
    parser = build_cli()
    args = parser.parse_args()
    message = cmd_to_str(args)
    if args.command == const.DOWNLOAD and args.data_source == const.SOURCE__DB:
        # Since the first element is the file, message[1] is the dataset type.
        print(message[0])
    else:
        print(message)



"""
Common constants used by the package.

We maintain this file to make sure that we don't have inconsistencies in the
constants used. It is common to notice spelling mistakes associated with strings
additionally, we don't get IDE's to automatically suggest values.

consider:

```python
a_dict["key"]
```

and 

```
a_dict[const.KEY]
```

In the latter case, a mature IDE will suggest the KEY constant, reducing time and ensuring consistency.
"""
TASK_TYPE__CONVERSATION = "conversation"
TASK_TYPE__SIMULATED_CALL = "simulated_call"
TASK_TYPE__AUDIO_SEGMENT = "audio_segment"
TASK_TYPE__DICT = "dict"
TASK_TYPE__CALL_TRANSCRIPTION = "call_transcription"
TASK_TYPE__DATA_GENERATION = "data_generation"

TASK_TYPES = [
    TASK_TYPE__CONVERSATION,
    TASK_TYPE__SIMULATED_CALL,
    TASK_TYPE__AUDIO_SEGMENT,
    TASK_TYPE__DICT,
    TASK_TYPE__CALL_TRANSCRIPTION,
    TASK_TYPE__DATA_GENERATION,
]

DOWNLOAD = "download"
UPLOAD = "upload"
DESCRIBE = "describe"
STATS = "stats"

OUTPUT_FORMAT__CSV = ".csv"
OUTPUT_FORMAT__SQLITE = ".sqlite"

SOURCE__DB = "tog"
SOURCE__DVC = "dvc"

DATASET_SERVER_URL = "DATASET_SERVER_URL"
DATA = "data"
DATA_ID = "data_id"
RAW = "raw"
CALL_UUID = "call_uuid"
CONVERSATION_UUID = "conversation_uuid"
UTTERANCES = "utterances"
ALTERNATIVES = "alternatives"
PRIORITY = "priority"
DATA_SOURCE = "data_source"
IS_GOLD = "is_gold"
DEFAULT_SOURCE = "calls"

UPLOAD_DATASET_SCHEMA = {
    "type": "object",
    "properties": {
        "state": {"type": "string"},
        "reftime": {"type": "string"},
        "audio_url": {
            "oneOf": [
                {"type": "string"},
                {
                    "type": "object",
                    "properties": {
                        "bucket": {"type": "string"},
                        "key": {"type": "string"},
                    },
                    "required": ["bucket", "key"],
                },
            ]
        },
        "call_uuid": {"type": "string"},
        "conversation_uuid": {"type": "string"},
        "alternatives": {"type": "array"},
        "filter": {
            "type": "object",
            "properties": {
                "predicted_intent": {"type": "string"},
                "confidence": {"type": "number"},
                "current_state": {"type": "string"},
                "expected_slots": {"type": "array"},
                "acknowledged_slots": {"type": "array"},
                "smalltalk": {"type": "boolean"},
            },
        },
    },
    "required": [
        "state",
        "reftime",
        "audio_url",
    ],
}

TOGDB_DB = "TOGDB_DB"
TOGDB_HOST = "TOGDB_HOST"
TOGDB_PORT = "TOGDB_PORT"
TOGDB_USER = "TOGDB_USER"
TOGDB_PASSWORD = "TOGDB_PASS"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['skit_labels']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.1,<4.0.0',
 'attrs>=19.3.0,<20.0.0',
 'docopt>=0.6.2,<0.7.0',
 'dvc[s3]>=2.9.3,<3.0.0',
 'jsonschema>=4.4.0,<5.0.0',
 'loguru>=0.5.3,<0.6.0',
 'psycopg2-binary>=2.8.5,<3.0.0',
 'pydash>=5.0.2,<6.0.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'pytz>=2020.1,<2021.0',
 'toml>=0.10.2,<0.11.0',
 'tqdm>=4.46.0,<5.0.0']

entry_points = \
{'console_scripts': ['skit-labels = skit_labels.cli:main']}

setup_kwargs = {
    'name': 'skit-labels',
    'version': '0.3.5',
    'description': 'Command line tool for interacting with labelled datasets at skit.ai.',
    'long_description': None,
    'author': None,
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import asyncio
import json
import os
import tempfile
from typing import Iterable, List, Optional, Tuple, Union

import aiohttp
import attr
import dvc.api
import jsonschema
import numpy as np
import pandas as pd
import pytz
from loguru import logger
from tqdm import tqdm

from skit_labels import constants as const
from skit_labels.db import Job, SqliteDatabase


def batch_gen(source, n=100):
    """
    Batched generator
    """

    batch = []

    for it in source:
        batch.append(it)
        if len(batch) % n == 0:
            yield batch
            batch = []

    if batch:
        yield batch


def download_dataset(
    job_id: str,
    task_type: str,
    timezone: pytz.BaseTzInfo = pytz.UTC,
    full: bool = False,
    batch_size: int = 500,
    start_date: Optional[str] = None,
    end_date: Optional[str] = None,
    db: Optional[str] = None,
    user: Optional[str] = None,
    password: Optional[str] = None,
    host: Optional[str] = None,
    port: Optional[Union[int, str]] = None,
) -> Tuple[SqliteDatabase, str, str]:
    job = Job(
        int(job_id),
        task_type=task_type,
        tz=timezone,
        db=db,
        user=user,
        password=password,
        host=host,
        port=port,
    )
    describe_dataset(job_id)
    stat_dataset(job_id)

    _, temp_filepath = tempfile.mkstemp(suffix=const.OUTPUT_FORMAT__SQLITE)
    sdb = SqliteDatabase(temp_filepath)
    bar = tqdm(total=job.total(untagged=full))

    for items in batch_gen(
        job.get(untagged=full, start_date=start_date, end_date=end_date),
        n=int(batch_size),
    ):
        rows = []
        for task, tag, tagged_time in items:
            # For raw dictionary type tasks, we don't use attr classes.
            if isinstance(task, dict):
                task_dict = task
            else:
                task_dict = attr.asdict(task)

            # TODO: is_gold might not be working for dict type tasks as of now
            rows.append((task.id, task_dict, tag, task.is_gold, tagged_time))

        sdb.insert_rows(rows)
        bar.update(n=len(items))
    return sdb, temp_filepath, job.type()


def sdb2df(sdb: SqliteDatabase, job_id: str) -> str:
    _, output_file = tempfile.mkstemp(
        prefix=f"job-{job_id}-", suffix=const.OUTPUT_FORMAT__CSV
    )
    df = pd.read_sql_query("SELECT * FROM data", sdb.conn)
    df.to_csv(output_file, index=False)
    return output_file


def describe_dataset(
    job_id: Optional[int] = None,
    job: Optional[Job] = None,
    db: Optional[str] = None,    
    user: Optional[str] = None,
    password: Optional[str] = None,
    host: Optional[str] = None,
    port: Optional[Union[int, str]] = None,
) -> str:
    job_ = job or Job(
        int(job_id),
        db=db,
        user=user,
        password=password,
        host=host,
        port=port,
    )

    return f"Job {job_.id}: {job_.name} [language: {job_.lang}]\n{job_.description}"


def stat_dataset(
    job_id: Optional[int] = None,
    job: Optional[Job] = None,
    db: Optional[str] = None,
    user: Optional[str] = None,
    password: Optional[str] = None,
    host: Optional[str] = None,
    port: Optional[Union[int, str]] = None,
) -> str:
    job_ = job or Job(
        int(job_id),
        db=db,
        user=user,
        password=password,
        host=host,
        port=port,
    )
    n_total = job_.total(untagged=True)
    n_tagged = job_.total()
    return f"Total items {n_total}. Tagged {n_tagged}. Untagged {n_total - n_tagged}."


def download_dataset_from_dvc(
    repo: str, path: str, remote: Optional[str] = None
) -> str:
    file_name = os.path.split(path)[-1]
    _, output_file = tempfile.mkstemp(suffix=file_name)
    with dvc.api.open(path, repo=repo, remote=remote) as f:
        df = pd.read_csv(f)
        df.to_csv(output_file, index=False)
    return output_file


def download_dataset_from_db(
    job_id: str,
    task_type: str,
    timezone: pytz.BaseTzInfo = pytz.UTC,
    full: bool = False,
    batch_size: int = 500,
    output_format: str = const.OUTPUT_FORMAT__CSV,
    start_date: Optional[str] = None,
    end_date: Optional[str] = None,
    db: Optional[str] = None,
    user: Optional[str] = None,
    password: Optional[str] = None,
    host: Optional[str] = None,
    port: Optional[Union[int, str]] = None,
) -> Tuple[str, str]:
    sdb, sdb_path, dataset_type = download_dataset(
        job_id,
        task_type,
        timezone,
        full,
        batch_size,
        start_date=start_date,
        end_date=end_date,
        db=db,
        user=user,
        password=password,
        host=host,
        port=port,
    )
    if output_format == const.OUTPUT_FORMAT__CSV:
        df_path = sdb2df(sdb, job_id)
        os.remove(sdb_path)
        return df_path, dataset_type
    else:
        return sdb_path, dataset_type


def build_dataset(
    job_id: str, data_frame: pd.DataFrame, source: Optional[str] = const.DEFAULT_SOURCE
) -> List[dict]:
    """
    Build a dataset from the dataframe.

    :param job_id: The dataset id where data should be uploaded.
    :type job_id: str
    :param data_frame: The dataframe to upload.
    :type data_frame: pd.DataFrame
    :param source: The source of the data_frame.
    :type source: Optional[str]
    :return: The job-id where the data was uploaded.
    :rtype: int
    """
    dataset = []
    logger.debug(f"Pushing {len(data_frame)} items to {job_id=}")
    data_frame.fillna(np.nan, inplace=True)
    data_frame.replace([np.nan], [None], inplace=True)
    for _, row in tqdm(
        data_frame.iterrows(),
        total=len(data_frame),
        desc="Building a dataset for uploading safely.",
    ):
        dedupe_id = "_".join([row[const.CONVERSATION_UUID], row[const.CALL_UUID]])

        if const.RAW in data_frame.columns:
            data = json.loads(row[const.RAW])
        else:
            data = row.to_dict()

        alternatives = (
            data[const.UTTERANCES]
            if const.UTTERANCES in data
            else data[const.ALTERNATIVES]
        )
        if not alternatives:
            alternatives = []

        if isinstance(alternatives, str):
            alternatives = json.loads(alternatives)

        data_point = {
            const.PRIORITY: 1,
            const.DATA_SOURCE: source,
            const.DATA_ID: dedupe_id,
            const.DATA: {
                **data,
                const.CALL_UUID: str(row[const.CALL_UUID]),
                const.CONVERSATION_UUID: str(row[const.CONVERSATION_UUID]),
                const.ALTERNATIVES: alternatives,
            },
            const.IS_GOLD: False,
        }
        jsonschema.validate(data_point[const.DATA], const.UPLOAD_DATASET_SCHEMA)
        dataset.append(data_point)
    return dataset


async def upload_dataset(
    session: aiohttp.ClientSession, job_id: str, dataset: List[dict]
):
    path = f"/tog/tasks/?job_id={job_id}"
    async with session.post(path, json=dataset) as response:
        upload_response = await response.json()
        return (upload_response, response.status)


async def upload_dataset_batches(
    dataset_chunks: Iterable[List[dict]], url: str, token: str, job_id: str
) -> str:
    """
    Post the dataset to the server.

    :param dataset_chunks: The dataset to post.
    :type dataset_chunks: Iterable[dict]
    :param token: The token to use for authentication.
    :type token: str
    :param job_id: The job id where the dataset should be uploaded.
    :type job_id: str
    :return: The job id where the dataset was uploaded.
    :rtype: int
    """
    headers = {"Authorization": f"Bearer {token}"}
    async with aiohttp.ClientSession(url, headers=headers) as session:
        requests = [
            upload_dataset(session, job_id, dataset) for dataset in dataset_chunks
        ]
        return await asyncio.gather(*requests)


async def upload_dataset_to_db(
    input_file: str, url: str, token: str, job_id: str
) -> Tuple[List[str], int]:
    """
    Uploads a dataset to the database.

    :param input_file: Path to the input file.
    :type input_file: str
    :param url: The url to the dataset server.
    :type url: str
    :param token: The token for uploading to the target organization.
    :type token: str
    :param job_id: The dataset id where data should be uploaded, defaults to None
    :type job_id: Optional[str], optional
    :return: The job-id where the data was uploaded.
    :rtype: str
    """
    _, extension = os.path.splitext(input_file)

    if extension != ".csv":
        raise ValueError("Expected file extension to be a csv.")

    data_frame = pd.read_csv(input_file)
    dataset = build_dataset(job_id, data_frame, token)
    batched_dataset = batch_gen(dataset)
    responses = await upload_dataset_batches(batched_dataset, url, token, job_id)
    errors = []

    for message, status_code in responses:
        if status_code != 200:
            errors.append(message)
            logger.error(message)
    return errors, len(data_frame)



"""
Module provides access to logger config, session token and package version.
"""
import os
import sys
from typing import Optional

import toml
from loguru import logger

LOG_LEVELS = ["CRITICAL", "ERROR", "WARNING", "SUCCESS", "INFO", "DEBUG", "TRACE"]


def get_version():
    project_toml = os.path.abspath(
        os.path.join(os.path.dirname(__file__), "..", "pyproject.toml")
    )
    with open(project_toml, "r") as handle:
        project_metadata = toml.load(handle)
    return project_metadata["tool"]["poetry"]["version"]


def configure_logger(level: int) -> None:
    """
    Configure the logger.
    """
    size = len(LOG_LEVELS)
    if level >= size:
        level = size - 1
    log_level = LOG_LEVELS[level]

    config = {
        "handlers": [
            {
                "sink": sys.stdout,
                "format": """
    -------------------------------------------------------
    <level>{level}</level>
    -------
    TIME: <green>{time}</green>
    FILE: {name}:L{line} <blue>{function}(...)</blue>
    <level>{message}</level>
    -------------------------------------------------------
    """,
                "colorize": True,
                "level": log_level,
            },
            {
                "sink": "file.log",
                "rotation": "500MB",
                "retention": "10 days",
                "format": "{time} {level} -\n{message}\n--------------------\n",
                "level": log_level,
            },
        ],
    }
    logger.configure(**config)
    logger.enable(__name__)


def read_session() -> Optional[str]:
    """
    Read the session from the environment.
    """
    home = os.path.expanduser("~")
    try:
        with open(os.path.join(home, ".skit", "token"), "r") as handle:
            return handle.read().strip()
    except FileNotFoundError:
        return None
